#include "fill.h"

#include <stdlib.h>
#include <conio.h>     // For interrupt() only
#include <mem.h>

enum { DN, LEFT, RIGHT, L_DN, R_DN, UP, L_UP, R_UP, NO };

///////////////////////////////////////////////////////////////////////////
inline int FillArea::get_pixel(loc p)                     // is this pixel
    {                                                     // have color ==
    if(getpixel(p) != start_color || !bound.contains(p))  // start_color,
	return 0;                                         // and is it inside
    return 1;                                             // of bound (work
    }                                                     // area

inline int FillArea::get_pixel(int x, int y) { return get_pixel(loc(x, y)); }
////////////////////////////
int FillArea::inside(loc p)      // is there any line under current pixel ?
    {
    for(int i = 0; i < used; i++)
	{
	if(p.Y > level[i].origin.Y - 1)
	    continue;
	else if(p.Y < level[i].origin.Y - 1)
	    return 0;
	if(p.X <= level[i].corner.X && p.X >= level[i].origin.X)
	    return 1;
	}
    return 0;
    }
///////////////////////
int FillArea::go_local(int left)     // go to the local minimum under the current
    {                          // pixel. "left" is used to block left moving
    int i = 1;                 // if we use the same start and fill colors
    loc pos = getCP();         // and can not begin from step down
    while(1)
	{
	if(get_pixel(pos.X, pos.Y + 1) && !inside(pos))
	    { moveto(pos = loc(pos.X, pos.Y + 1)); i = 0; } // left = 1; }
	else if(left && get_pixel(pos.X - 1, pos.Y))
	    { moveto(pos = loc(pos.X - 1, pos.Y)); i = 0; }
	else break;
	}
    local = 1;
    return i;
    }
////////////////////////
rect FillArea::get_line()                // return the line under current
    {                                    // pixel
    loc pos = getCP();
    rect line_below = rect(pos, pos);
    for(int i = 0; i <= used; i++)
	{
	if(level[i].corner.Y == pos.Y && level[i].contains(pos))
	    {
	    line_below = level[i];
	    break;
	    }
	}
    return line_below;
    }
///////////////////////
void FillArea::remove(rect r)                     // remove line from array
    {
    for(int i = 0; i <= used; i++)
	{
	if(level[i].origin.Y == r.origin.Y
	   && level[i].origin.X == r.origin.X)
	    {
	    for(; i <= used; i++)
		level[i] = level[i + 1];
	    used--;
	    break;
	    }
	}
    }
///////////////////////
void FillArea::add(rect r)       // add line to array. Lines are sorted
    {                            // by Y, then by X
    int i;
    for(i = 0; i <= used; i++)
	{
	if(level[i].origin.Y == r.origin.Y && level[i].origin.X > r.origin.X)
	    break;
	else if(level[i].origin.Y > r.origin.Y)
	    break;
	}
    for(int j = used; j >= i; j--)
	level[j + 1] = level[j];
    if(i > used)
	i--;
    level[i] = r;
    work = r;
    used++;
    }
///////////////////////////
void FillArea::draw_line()  // draws line with current colors and pattern
    {
    flag = 1;
    loc pos = getCP();
    rect line_below = get_line();          // if we draw line above any other
    rect new_rect;                         // contains new line
    remove(line_below);
    loc p = getCP();
    if(!local)                            // it is not local minimum
	{
	moveto(loc(line_below.origin.X, line_below.origin.Y - 1));   // go up
	while(1)
	    {
	    p = getCP();
	    if((!get_pixel(p))               // no place above leftmost pixel
		&& line_below.corner.X >= p.X)       // and we can move right
		{
		if(line_below.corner.X == p.X)       // if it is the end of
		    {                                // line_below
		    if(!used)                        // if no ways from this
			flag = 0;                    // point
		    break;
		    }
		else moveto(loc(p.X + 1, p.Y));
		}
	    else break;
	    }

	while(1)
	    {
	    p = getCP();
	    if((get_pixel(p.X - 1, p.Y))                  // left, but not dn
		&& get_pixel(p)
		&& (!get_pixel(p.X, p.Y + 1)
		    || p.X >= line_below.origin.X)        // or we are above
		&& !inside(loc(p.X - 1, p.Y - 1)))        // and no intersection
		moveto(p.X - 1, p.Y);                     // with other lines
	    else
		{
		if(line_below.origin.X > p.X              // we can go_local
		    && get_pixel(p.X, p.Y + 1))
		    flag = 0;
		break;
		}
	    }
	}
    int mark = 0;
    if(get_pixel(p))            // first pixel of line
	{
	put_pixel(p.X, p.Y);
	mark = 1;               // mark that at least one pixel drawn
	}
    new_rect = rect(p, p);

    int disable = 1;           // first time impossible to put pixel

    while(1)
	{
	p = loc(p.X + 1, p.Y);
	if( mark
	    && get_pixel(p)
	    && (p.X <= line_below.corner.X
		|| (!get_pixel(p.X, p.Y + 1)
		    || inside(p))))
	    {
	    put_pixel(p.X, p.Y);        // draw line in cycle
	    if(disable)
		new_rect.corner = p;
	    else
		new_rect.origin = p;
	    disable = 1;
	    }
	else
	    {
	    if(disable && mark)          // all with this line
		add(new_rect);
	    if(get_pixel(p.X, p.Y + 1)   // we can go_local
		&& get_pixel(p))
		flag = 0;
	    if(p.X >= line_below.corner.X)
		break;
	    else
		new_rect = rect(p, loc(line_below.corner.X,
				       line_below.corner.Y - 1));
	    disable = 0;
	    }
	moveto(p);
	}
    local = 0;
    }
////////////////////////
void FillArea::dump_levels()                          // simplify the level[]
    {
    for(int i = 0; i < used; i++)
	for(int j = i + 1; j < used; j++)
	    {
	    if(level[i].origin.X - 1 == level[j].corner.X
	       && level[i].origin.Y == level[j].origin.Y)
	       {
	       level[i].origin.X = level[j].origin.X;
	       remove(level[j]);
	       }
	    if(level[i].corner.X + 1 == level[j].origin.X
	       && level[i].origin.Y == level[j].origin.Y)
	       {
	       level[i].corner.X = level[j].corner.X;
	       remove(level[j]);
	       }

	    if(level[i].corner.X >= level[j].origin.X
	       && level[i].origin.X <= level[j].corner.X
	       && level[i].origin.Y == level[j].origin.Y - 1)
		{
		if(level[i].origin.X <= level[j].origin.X)
		    level[j].origin.X = level[i].corner.X;
		else
		    level[j].corner.X = level[i].origin.X;
		if(level[j].origin.X >= level[j].corner.X)
		    remove(level[j]);
		}
	    }
    }
////////////////////////
void FillArea::fill_cell()        // fill before it is possible to go_local
    {
    while(flag && used && !interrupted())
	{
	loc pos = getCP();

	if(pos.Y == 75)
	    pos.Y = 75;

	moveto(level[used - 1].corner);
	draw_line();
	dump_levels();
	}
    if(interrupted())
	for(int i = 0; i < 640; i++)
	    level[i] = rect(0, 0, 0, 0);
    }
////////////////////////
void FillArea::put_pixel(int x, int y)
    {
    if(fill_style == PATTERN_FILL)
	{
	int sx, sy;
	sx = x % 8;
	sy = y % 8;
	int col = pattern[pat][sy] & (1 << sx) ? bak : attr;
	::putpixel(x, y, col);
	}
    else
	{
	if(random(16) > pat)
	    putpixel(x, y, attr);
	else
	    putpixel(x, y, bak);
	}
    }
//////////////////////////////////
void FillArea::fill_area(loc from, rect r)
    {
    memset(level, 0, 640 * sizeof(rect));
    if(!r.contains(from))
	return;
    flag = 1;
    moveto(from);
    start_color = getpixel(from);
    bound = r;

    go_local();          // go to start position
    draw_line();         // draw the first line and add first line to level[]

    while(1)
	{
	if(flag)
	    fill_cell();
	flag = 1;
	if(!used)
	    return;
	moveto(work.origin);
	if(go_local())     // left local min
	    {
	    moveto(work.corner.X + 1, work.corner.Y);
	    if(go_local(0))    // right local min
		return;
	    }
	draw_line();
	}
    }
////////////////////////
int FillArea::interrupted()
    {
    return kbhit();
    }

void main()
    {
    int gdriver = DETECT, gmode;
    initgraph(&gdriver, &gmode, "c:\\borlandc\\BGI");

    FillArea* fill = new FillArea();
    fill->flag = 0;
    fill->used = 0;

    circle(120, 120, 100);
    line(80, 80, 100, 100);
    line(100, 100, 120, 100);
    line(120, 100, 140, 75);

    loc from(125, 125);
    fill->pat = 12;
    fill->fill_style = PATTERN_FILL;
    fill->attr = GREEN;
    fill->bak = RED;

    fill->fill_area(from); //, r);

    delete fill;
    closegraph();
    }
